use std::{
    fs::File,
    io::{BufReader, Read},
};

use eframe::egui;
use error_stack::{Report, ResultExt};
use tracing::error;

use crate::{
    application::{errors::FileError, Application},
    excellon::{doc::ExcellonDoc, drills::Drills, parse_excellon},
    geometry::Geometry,
    gerber::{doc::GerberDoc, parse_gerber},
};

const GERBER_EXTENSIONS: &[&str] = &["GBR ", "gbr", "GB", "geb"];
const EXCELLON_EXTENSIONS: &[&str] = &["DRL ", "drl"];

pub fn draw_header(ctx: &egui::Context, app: &mut Application) {
    egui::TopBottomPanel::top("top_panel")
        .exact_height(40.)
        .show(ctx, |ui| {
            ui.horizontal_centered(|ui| {
                if let Some(opened_files) = ui
                    .menu_button("📝 Open File", open_file_menu)
                    .inner
                    .flatten()
                {
                    match opened_files {
                        Ok(files) => {
                            for (path, name, content) in files {
                                match content {
                                    AppFileContent::Gerber(gerber) => {
                                        app.gerbers.insert(
                                            path,
                                            (
                                                name,
                                                Geometry::from(gerber)
                                                    .into_unit(app.variables.units),
                                            ),
                                        );
                                    }
                                    AppFileContent::Excellon(excellon) => {
                                        let drills: Drills = excellon.into();
                                        app.excellons.insert(
                                            path,
                                            (name, drills.into_unit(app.variables.units)),
                                        );
                                    }
                                }
                            }
                        }
                        Err(e) => {
                            // TODO Show error
                            error!("{e:?}");
                        }
                    }
                };
            })
        });
}

type FileOpenResponse = Result<Vec<(String, String, AppFileContent)>, Report<FileError>>;

fn open_file_menu(ui: &mut egui::Ui) -> Option<FileOpenResponse> {
    let mut response = None;
    ui.set_max_width(200.0); // To make sure we wrap long text

    if ui.button("Gerber").clicked() {
        response = Some(open_file(AppFileType::Gerber));
        ui.close_menu();
    }

    if ui.button("Excellon").clicked() {
        ui.close_menu();
        response = Some(open_file(AppFileType::Excellon));
    }

    // ui.menu_button("SubMenu", |ui| {
    //     ui.menu_button("SubMenu", |ui| {
    //         if ui.button("Open…").clicked() {
    //             ui.close_menu();
    //         }
    //         let _ = ui.button("Item");
    //     });
    //     ui.menu_button("SubMenu", |ui| {
    //         if ui.button("Open…").clicked() {
    //             ui.close_menu();
    //         }
    //         let _ = ui.button("Item");
    //     });
    //     let _ = ui.button("Item");
    //     if ui.button("Open…").clicked() {
    //         ui.close_menu();
    //     }
    // });
    // ui.menu_button("SubMenu", |ui| {
    //     let _ = ui.button("Item1");
    //     let _ = ui.button("Item2");
    //     let _ = ui.button("Item3");
    //     let _ = ui.button("Item4");
    //     if ui.button("Open…").clicked() {
    //         ui.close_menu();
    //     }
    // });
    // let _ = ui.button("Very long text for this item that should be wrapped");

    response
}

#[derive(Debug, Clone, Copy)]
enum AppFileType {
    Gerber,
    Excellon,
}

impl std::fmt::Display for AppFileType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                AppFileType::Gerber => "Gerber",
                AppFileType::Excellon => "Excellon",
            }
        )
    }
}

impl AppFileType {
    pub fn extensions(&self) -> &[&str] {
        match self {
            AppFileType::Gerber => GERBER_EXTENSIONS,
            AppFileType::Excellon => EXCELLON_EXTENSIONS,
        }
    }
}

pub enum AppFileContent {
    Gerber(GerberDoc),
    Excellon(ExcellonDoc),
}

impl AppFileContent {
    fn from_file<T: Read>(
        file_type: AppFileType,
        file: BufReader<T>,
    ) -> Result<Self, Report<FileError>> {
        Ok(match file_type {
            AppFileType::Gerber => Self::Gerber(parse_gerber(file)),
            AppFileType::Excellon => Self::Excellon(
                parse_excellon(file)
                    .change_context(FileError)
                    .attach_printable("Could not parse Excellon file")?,
            ),
        })
    }
}

fn open_file(
    file_type: AppFileType,
) -> Result<Vec<(String, String, AppFileContent)>, Report<FileError>> {
    let mut result = Vec::new();
    if let Some(paths) = rfd::FileDialog::new()
        .add_filter(file_type.to_string(), file_type.extensions())
        .pick_files()
    {
        for path in paths {
            // get file content
            let file = File::open(&path).change_context(FileError)?;
            // parse file format
            let content = AppFileContent::from_file(file_type, BufReader::new(file))?;
            // get file path
            let path_str = path
                .to_str()
                .ok_or(FileError::new("Could not get file path string"))?;
            // get file name
            let name_str = path
                .file_name()
                .ok_or(FileError::new(
                    "Could not determine file name from selection",
                ))?
                .to_str()
                .ok_or(FileError::new("Could not convert OS String"))?;

            result.push((path_str.into(), name_str.into(), content));
        }
    }

    Ok(result)
}
