pub mod doc;
pub mod drills;
mod errors;

use std::{
    io::{BufRead, BufReader, Read},
    str::Chars,
};

use doc::{Command, ExcellonDoc, FunctionCode, Tool, Zeroes};
use error_stack::{Report, ResultExt};
use errors::ExcellonError;
use lazy_regex::{regex, Regex};
use tracing::{debug, error};

use crate::geometry::{point::Point, Unit};

pub fn parse_excellon<T: Read>(reader: BufReader<T>) -> Result<ExcellonDoc, Report<ExcellonError>> {
    let mut excellon_doc = ExcellonDoc::new();

    // Number format and units
    // INCH uses 6 digits
    // METRIC uses 5/6
    let re_units = regex!("^(INCH|METRIC)(?:,([TL])Z)?$");

    let re_toolset = regex!(r#"^T(\d+)(?:.*C(\d*\.?\d*))?"#);

    let re_coordinates = regex!(r#"X?(-?[0-9.]+)?Y?(-?[0-9.]+)?"#);

    // Parse coordinates
    let re_leading_zeroes = regex!(r#"^[-\+]?(0*)(\d*)"#);

    let mut inside_header = false;

    for (index, line) in reader.lines().enumerate() {
        let raw_line = line.change_context(ExcellonError)?;
        let line = raw_line.trim();

        // Show the line
        debug!("{}. {}", index + 1, &line);

        if !line.is_empty() {
            let mut line_chars = line.chars();

            if inside_header {
                match get_next_char(&mut line_chars)? {
                    'I' => parse_units(line, re_units, &mut excellon_doc),
                    'M' => match get_next_char(&mut line_chars)? {
                        'E' => parse_units(line, re_units, &mut excellon_doc),
                        '9' => inside_header = false, // End of header
                        _ => line_parse_failure(line, index),
                    },
                    'T' => parse_toolset(line, re_toolset, &mut excellon_doc)?,
                    '%' => inside_header = false, // End of header
                    ';' => parse_comment(line_chars, &mut excellon_doc),
                    _ => line_parse_failure(line, index),
                }
            } else {
                match get_next_char(&mut line_chars)? {
                    'G' => {}
                    'M' => match get_next_char(&mut line_chars)? {
                        '1' => {}
                        '3' => excellon_doc
                            .commands
                            .push(Command::FunctionCode(FunctionCode::EndOfFile)),
                        '4' => match get_next_char(&mut line_chars)? {
                            '8' => inside_header = true,
                            _ => line_parse_failure(line, index),
                        },
                        _ => line_parse_failure(line, index),
                    },
                    'T' => {
                        // Select Tool
                        let id = line_chars
                            .as_str()
                            .parse::<u32>()
                            .change_context(ExcellonError)?;
                        excellon_doc
                            .commands
                            .push(Command::FunctionCode(FunctionCode::SelectTool(id)))
                    }
                    'X' | 'Y' => {
                        if let Some(reg_match) = re_coordinates.captures(line) {
                            let x = reg_match
                                .get(1)
                                .ok_or(ExcellonError)
                                .attach_printable("No x coordinate present")?
                                .as_str();
                            let y = reg_match
                                .get(2)
                                .ok_or(ExcellonError)
                                .attach_printable("No y coordinate present")?
                                .as_str();

                            let x = if x.contains(".") {
                                x.parse::<f64>().change_context(ExcellonError)?
                            } else {
                                parse_number(x, &excellon_doc, re_leading_zeroes)?
                            };

                            let y = if y.contains(".") {
                                y.parse::<f64>().change_context(ExcellonError)?
                            } else {
                                parse_number(y, &excellon_doc, re_leading_zeroes)?
                            };

                            excellon_doc
                                .commands
                                .push(Command::FunctionCode(FunctionCode::Drill(Point::new(x, y))))
                        } else {
                            error!("Could not parse Coordinates")
                        };
                    }
                    ';' => parse_comment(line_chars, &mut excellon_doc),
                    _ => line_parse_failure(line, index),
                }
            }
        }
    }

    Ok(excellon_doc)
}

fn get_next_char(chars: &mut Chars) -> Result<char, ExcellonError> {
    chars.next().ok_or(ExcellonError)
}

// parse a Excellon Comment (e.g. '; This line is a comment and is ignored')
fn parse_comment(line: Chars, doc: &mut ExcellonDoc) {
    let comment = line.as_str();
    doc.commands
        .push(Command::FunctionCode(FunctionCode::Comment(
            comment.to_string(),
        )));
}

// parse a Excellon unit statement (e.g. 'METRIC,LZ')
fn parse_units(line: &str, re: &Regex, doc: &mut ExcellonDoc) {
    // Set the unit type
    if let Some(reg_match) = re.captures(line) {
        // Parse Unit
        doc.units = match reg_match.get(1).map(|s| s.as_str()) {
            Some("METRIC") => Unit::Millimeters,
            Some("INCH") => Unit::Inches,
            _ => {
                error!("Incorrect excellon units format");
                doc.units
            }
        };
        // Parse Zeroes Type (Leading | Trailing)
        doc.zeroes = match reg_match.get(2).map(|s| s.as_str()) {
            Some("L") => doc::Zeroes::Leading,
            Some("T") => doc::Zeroes::Trailing,
            _ => {
                error!("Incorrect excellon zeroes format");
                doc.zeroes
            }
        }
    }
}

fn parse_toolset(
    line: &str,
    re: &Regex,
    doc: &mut ExcellonDoc,
) -> Result<(), Report<ExcellonError>> {
    if let Some(reg_match) = re.captures(line) {
        if let (Some(id), Some(size)) = (
            reg_match.get(1).map(|s| s.as_str()),
            reg_match.get(2).map(|s| s.as_str()),
        ) {
            doc.tools.insert(
                id.parse::<u32>().change_context(ExcellonError)?,
                Tool::new(size.parse::<f64>().change_context(ExcellonError)?),
            );
        }
    }
    Ok(())
}

// print a simple message in case the parser hits a dead end
fn line_parse_failure(line: &str, index: usize) {
    error!(
        "## Excellon Parser ## Cannot parse line:\n{} | {}",
        index, line
    )
}

// Parses coordinate numbers without period.
fn parse_number(
    number_str: &str,
    doc: &ExcellonDoc,
    lz: &Regex,
) -> Result<f64, Report<ExcellonError>> {
    match doc.zeroes {
        Zeroes::Leading => {
            // With leading zeros, when you type in a coordinate,
            // the leading zeros must always be included.  Trailing zeros
            // are unneeded and may be left off. The CNC-7 will automatically add them.
            // r'^[-\+]?(0*)(\d*)'
            // 6 digits are divided by 10^4
            // If less than size digits, they are automatically added,
            // 5 digits then are divided by 10^3 and so on.

            let re_match = lz
                .captures(number_str)
                .ok_or(ExcellonError)
                .attach_printable("Leading zeroes regex does not match")?;
            let (g1, g2) = (
                re_match
                    .get(1)
                    .ok_or(ExcellonError)
                    .attach_printable("Regex MatchGroup 2")?
                    .as_str(),
                re_match
                    .get(2)
                    .ok_or(ExcellonError)
                    .attach_printable("Regex MatchGroup 2")?
                    .as_str(),
            );

            Ok(number_str.parse::<f64>().change_context(ExcellonError)?
                / (10_f64.powi(
                    g1.len() as i32 + g2.len() as i32
                        - match doc.units {
                            Unit::Inches => 2,
                            Unit::Millimeters => 3,
                        },
                )))
        }
        Zeroes::Trailing => {
            // Trailing
            // You must show all zeros to the right of the number and can omit
            // all zeros to the left of the number. The CNC-7 will count the number
            // of digits you typed and automatically fill in the missing zeros.
            Ok(number_str.parse::<f64>().change_context(ExcellonError)?
                / match doc.units {
                    Unit::Millimeters => 1000., // Metric is 000.000
                    Unit::Inches => 10000.,     // Inches is 00.0000
                })
        }
    }
}

#[cfg(test)]
mod tests {
    use std::fs::File;

    use super::*;

    #[test]
    fn parse_excellon_file() {
        let excellon = r#"; This line is a comment and is ignored 
            ; The next line starts the "header": 
            M48
            ; Units and number format:
            INCH,LZ
            ; One tool is defined with diameter 0.04 inches,
            ; drill rate of 300 inches/minute and 55000 RPM.
            T1C.04F300S55
            ; End of header, M95 or %, and beginning of body:
            M95

            ; Use tool 1 defined in the header
            T1
            ; Drill at points (123.45, 234.5) and (12.345, 234.5):
            X12345Y23450
            X012345Y234500
            ; End of program;
            M30"#;
        let file = File::open("./FirstPCB-PTH.drl").unwrap();
        let _reader = BufReader::new(file);

        let _excellon = parse_excellon(BufReader::new(excellon.as_bytes()));
    }
}
