use svg::{
    node::element::{path::Data, Circle, Path},
    Document,
};

use crate::{application::panels::actions::HOLE_MARK_DIAMETER, outline_geometry::OutlineGeometry};

pub struct SVGConverter;

impl SVGConverter {
    pub fn export(geo: &OutlineGeometry, file: &std::path::Path) {
        let view_box = geo.bounding_box;

        let data: Vec<Data> = geo
            .paths()
            .iter()
            .filter_map(|path| {
                if path.len() > 1 {
                    let mut iter = path.iter().map(|point| (point.x(), -point.y()));
                    let first = iter.next().unwrap();
                    let mut data = Data::new().move_to(first);
                    for point in iter {
                        data = data.line_to(point);
                    }

                    data = data.close();

                    Some(data)
                } else {
                    None
                }
            })
            .collect();

        let paths = data.into_iter().map(|data| {
            Path::new()
                .set("fill", "none")
                .set("stroke", "black")
                .set("stroke-width", geo.stroke)
                .set("d", data)
        });

        let points = geo.points().into_iter().map(|p| {
            Circle::new()
                .set("r", HOLE_MARK_DIAMETER)
                .set("cx", p.x())
                .set("cy", p.invert_y().y())
                .set("fill", "black")
        });

        let document_width = view_box.max.x() - view_box.min.x();
        let document_height = view_box.max.y() - view_box.min.y();

        let mut document = Document::new()
            .set(
                "viewBox",
                (
                    view_box.min.x(),
                    -view_box.min.y() - document_height,
                    document_width,
                    document_height,
                ),
            )
            .set("width", format!("{document_width}{}", geo.unit))
            .set("height", format!("{document_height}{}", geo.unit));

        for path in paths {
            document = document.add(path);
        }

        for point in points {
            document = document.add(point);
        }

        svg::save(file, &document).unwrap();
    }
}
