use std::collections::HashMap;

use clipper2::PointInPolygonResult;
use itertools::Itertools;

use crate::{
    application::CanvasColour,
    geometry::{point::Point, union::union_function, ClipperPath, ClipperPaths, DrawableRaw, Unit},
};

use super::Element;

#[derive(Debug, Clone)]
pub struct Composite {
    elements: Vec<Element>,
    outline: ClipperPaths,
}

impl Composite {
    pub fn new() -> Self {
        Self {
            elements: Vec::new(),
            outline: ClipperPaths::new(vec![]),
        }
    }

    pub fn add(&mut self, element: Element) {
        self.elements.push(element);
    }

    pub fn finalize(&mut self) {
        let mut intersection_map: HashMap<usize, Vec<usize>> = HashMap::new();

        for part in self.elements.iter().enumerate().combinations(2) {
            let (index1, element1) = part[0];
            let (index2, element2) = part[1];

            // check if element1 and element2 intersect
            if element1
                .outline()
                .iter()
                .any(|p| element2.outline().is_point_inside(*p) != PointInPolygonResult::IsOutside)
            {
                // add intersection result for both indices
                let entry1 = intersection_map.entry(index1).or_default();
                entry1.push(index2);
                let entry2 = intersection_map.entry(index2).or_default();
                entry2.push(index1);
            }
        }

        // go through all aperture components
        for i in 0..intersection_map.len() {
            // remove component from map
            if let Some(mut intersections) = intersection_map.remove(&i) {
                // take outline of element as base
                let mut geo = self.elements[i].to_paths();

                // union with intersecting aperture components until done
                while let Some(other) = intersections.pop() {
                    // union with intersecting aperture component
                    let intersecting_element = &self.elements[other];
                    if let Some(union) = union_function(&geo, &intersecting_element.to_paths()) {
                        geo = union;
                    }

                    // get intersections of united aperture component and add them if not already in own
                    if let Some(new_intersections) = intersection_map.remove(&other) {
                        for o in new_intersections {
                            // add to original line if not self and not already inside
                            if o != i   // ensure to not intersect with itself
                            && !intersections.contains(&o)  // do not add if already in intersection list
                            && intersection_map.contains_key(&o)
                            // check if intersection was already performed
                            {
                                intersections.push(o);
                            }
                        }
                    }
                }
                self.outline.push(geo);
            }
        }
    }

    pub fn into_unit(self, origin: Unit, to: Unit) -> Self {
        Self {
            elements: self
                .elements
                .into_iter()
                .map(|el| el.into_unit(origin, to))
                .collect(),
            outline: self.outline,
        }
    }
}

impl DrawableRaw for Composite {
    fn canvas_pos(&self) -> Point {
        // self.position
        todo!()
    }

    fn draw_egui_plot(&self, ui: &mut egui_plot::PlotUi, colour: CanvasColour, selected: bool) {
        for element in &self.elements {
            element.draw_egui_plot(ui, colour, selected);
        }
    }

    fn to_paths(&self) -> ClipperPaths {
        self.outline.clone()
    }

    fn outline(&self) -> ClipperPath {
        self.outline.first().unwrap().clone()
    }
}
