use circle::Circle;
use composite::Composite;
use egui_plot::PlotUi;
use evalexpr::eval_number;
use gerber_types::MacroDecimal;
use lazy_regex::regex;
use linepath::LinePath;
use obround::Obround;
use polygon::Polygon;
use rectangle::Rectangle;
use tracing::error;

use crate::application::CanvasColour;

use super::{point::Point, ClipperPath, ClipperPaths, DrawableRaw, Unit};

pub mod circle;
pub mod composite;
pub mod linepath;
pub mod obround;
pub mod polygon;
pub mod rectangle;

#[derive(Debug, Clone)]
pub enum Element {
    Circle(Circle),
    Composite(Composite),
    Rectangle(Rectangle),
    _Line(LinePath),
    Obround(Obround),
    Polygon(Polygon),
}

impl Element {
    pub fn draw_egui_plot(&self, ui: &mut PlotUi, colour: CanvasColour, selected: bool) {
        match self {
            Element::Circle(c) => c.draw_egui_plot(ui, colour, selected),
            Element::Composite(co) => co.draw_egui_plot(ui, colour, selected),
            Element::Rectangle(r) => r.draw_egui_plot(ui, colour, selected),
            Element::_Line(l) => l.draw_egui_plot(ui, colour, selected),
            Element::Obround(o) => o.draw_egui_plot(ui, colour, selected),
            Element::Polygon(p) => p.draw_egui_plot(ui, colour, selected),
        }
    }

    pub fn canvas_pos(&self) -> Point {
        match self {
            Element::Circle(c) => c.canvas_pos(),
            Element::Composite(c) => c.canvas_pos(),
            Element::Rectangle(r) => r.canvas_pos(),
            Element::_Line(l) => l.canvas_pos(),
            Element::Obround(o) => o.canvas_pos(),
            Element::Polygon(p) => p.canvas_pos(),
        }
    }

    pub fn to_paths(&self) -> ClipperPaths {
        match self {
            Element::Circle(c) => c.to_paths(),
            Element::Composite(c) => c.to_paths(),
            Element::Rectangle(r) => r.to_paths(),
            Element::_Line(l) => l.to_paths(),
            Element::Obround(o) => o.to_paths(),
            Element::Polygon(p) => p.to_paths(),
        }
    }

    pub fn outline(&self) -> ClipperPath {
        match self {
            Element::Circle(c) => c.outline(),
            Element::Composite(c) => c.outline(),
            Element::Rectangle(r) => r.outline(),
            Element::_Line(l) => l.outline(),
            Element::Obround(o) => o.outline(),
            Element::Polygon(p) => p.outline(),
        }
    }

    pub fn into_unit(self, origin: Unit, to: Unit) -> Self {
        match self {
            Element::Circle(c) => Element::Circle(c.into_unit(origin, to)),
            Element::Composite(c) => Element::Composite(c.into_unit(origin, to)),
            Element::Rectangle(r) => Element::Rectangle(r.into_unit(origin, to)),
            Element::_Line(l) => Element::_Line(l.into_unit(origin, to)),
            Element::Obround(o) => Element::Obround(o.into_unit(origin, to)),
            Element::Polygon(p) => Element::Polygon(p.into_unit(origin, to)),
        }
    }
}

pub fn macro_decimal_to_f64(macro_decimal: &MacroDecimal, variables: &[f64]) -> f64 {
    let re_units = regex!(r#"(\$\d+)"#);

    match macro_decimal {
        MacroDecimal::Value(v) => *v,
        MacroDecimal::Variable(v) => *variables.get(*v as usize).unwrap_or(&0.),
        MacroDecimal::Expression(ex) => {
            // search for variables (eg. $1)
            if let Some(captures) = re_units.captures(ex) {
                // copy expression
                let mut expr = ex.clone();
                // go through all found variable names
                for c in captures.iter().skip(1).flatten() {
                    // get variable name
                    let matched = c.as_str();
                    // create index out of variable
                    let index = matched.replace('$', "").parse::<usize>().unwrap();
                    // replace variable name with value
                    expr = expr.replace(
                        matched,
                        &variables.get(index - 1).unwrap_or(&0.).to_string(),
                    );
                }
                match eval_number(&expr) {
                    Ok(res) => res,
                    Err(e) => {
                        error!("{e:?}");
                        0.
                    }
                }
            } else {
                0.
            }
        }
    }
}
