use eframe::egui::Stroke;
use egui_plot::{PlotPoints, Polygon};

use crate::{
    application::CanvasColour,
    geometry::{ClipperPath, ClipperPaths},
};

use super::super::{
    helpers::{create_circular_path, semi_circle, CircleSegment},
    point::{convert_to_unit, Point},
    DrawableRaw, Unit,
};
use super::rectangle::Rectangle;

#[derive(Debug, Clone)]
pub struct Obround {
    position: Point,
    x: f64,
    y: f64,
    _rounding: f64,
    outline: ClipperPath,
    _rectangle: Rectangle,
    hole_diameter: Option<f64>,
}

impl Obround {
    pub fn new(position: Point, x: f64, y: f64, hole_diameter: Option<f64>) -> Self {
        let diameter = if x < y { x } else { y };
        let outline = if x == y {
            create_circular_path(&position, x, CircleSegment::Full)
        } else {
            let mut path: Vec<(f64, f64)> = Vec::new();
            // check if obround is round to x or y
            if x < y {
                // round on y axis
                path.append(&mut semi_circle(
                    position - Point::new(0., y / 2. - x / 2.),
                    x,
                    180.,
                ));
                path.append(&mut semi_circle(
                    position + Point::new(0., y / 2. - x / 2.),
                    x,
                    0.,
                ));
            } else {
                path.append(&mut semi_circle(
                    position + Point::new(x / 2. - y / 2., 0.),
                    y,
                    270.,
                ));
                path.append(&mut semi_circle(
                    position - Point::new(x / 2. - y / 2., 0.),
                    y,
                    90.,
                ));
            }

            path
        };

        Self {
            position,
            x,
            y,
            _rounding: diameter,
            outline: outline.into(),
            _rectangle: Rectangle::new(position, x, y, hole_diameter),
            hole_diameter,
        }
    }

    pub fn from_aperture_obround(aperture: &gerber_types::Rectangular, position: Point) -> Self {
        Self::new(position, aperture.x, aperture.y, aperture.hole_diameter)
    }

    pub fn into_unit(self, origin: Unit, to: Unit) -> Self {
        Self::new(
            self.position.into_unit(origin, to),
            convert_to_unit(self.x, origin, to),
            convert_to_unit(self.y, origin, to),
            self.hole_diameter.map(|d| convert_to_unit(d, origin, to)),
        )
    }
}

impl DrawableRaw for Obround {
    fn canvas_pos(&self) -> Point {
        self.position
    }

    fn draw_egui_plot(&self, ui: &mut egui_plot::PlotUi, colour: CanvasColour, selected: bool) {
        let mut points: Vec<[f64; 2]> = self.outline.iter().map(|p| [p.x(), p.y()]).collect();
        points.push(points[0]);

        let poly = Polygon::new(PlotPoints::from(points))
            .fill_color(colour.as_colour32(selected))
            .stroke(Stroke::NONE);
        ui.polygon(poly);
    }

    fn to_paths(&self) -> ClipperPaths {
        ClipperPaths::new(vec![self.outline.clone()])
    }

    fn outline(&self) -> ClipperPath {
        self.outline.clone()
    }
}
