use eframe::egui::Stroke;
use egui_plot::{PlotPoints, Polygon};
use gerber_types::{CenterLinePrimitive, VectorLinePrimitive};

use crate::{
    application::CanvasColour,
    geometry::{ClipperPath, ClipperPaths},
};

use super::{
    super::{
        point::{convert_to_unit, Point},
        DrawableRaw, Unit,
    },
    macro_decimal_to_f64,
};

#[derive(Debug, Clone)]
pub struct Rectangle {
    position: Point,
    pub width: f64,
    pub height: f64,
    hole_diameter: Option<f64>,
    outline: ClipperPath,
}

impl Rectangle {
    pub fn from_aperture_rectangular(
        aperture: &gerber_types::Rectangular,
        position: Point,
    ) -> Self {
        Self {
            position,
            width: aperture.x,
            height: aperture.y,
            hole_diameter: aperture.hole_diameter,
            outline: vec![
                (
                    position.x() - aperture.x / 2.,
                    position.y() - aperture.y / 2.,
                ),
                (
                    position.x() - aperture.x / 2.,
                    position.y() + aperture.y / 2.,
                ),
                (
                    position.x() + aperture.x / 2.,
                    position.y() + aperture.y / 2.,
                ),
                (
                    position.x() + aperture.x / 2.,
                    position.y() - aperture.y / 2.,
                ),
                (
                    position.x() - aperture.x / 2.,
                    position.y() - aperture.y / 2.,
                ),
            ]
            .into(),
        }
    }

    pub fn new(position: Point, width: f64, height: f64, hole_diameter: Option<f64>) -> Self {
        Self {
            position,
            width,
            height,
            hole_diameter,
            outline: vec![
                (position.x() - width / 2., position.y() - height / 2.),
                (position.x() - width / 2., position.y() + height / 2.),
                (position.x() + width / 2., position.y() + height / 2.),
                (position.x() + width / 2., position.y() - height / 2.),
            ]
            .into(),
        }
    }

    pub fn from_macro_vector_line(
        ap_macro: &VectorLinePrimitive,
        variables: &[f64],
        target: Point,
    ) -> Self {
        // TODO angle
        let x_start = macro_decimal_to_f64(&ap_macro.start.0, variables);
        let x_end = macro_decimal_to_f64(&ap_macro.end.0, variables);
        let y_start = macro_decimal_to_f64(&ap_macro.start.1, variables);
        let y_end = macro_decimal_to_f64(&ap_macro.end.1, variables);

        let height = if x_end - x_start == 0. {
            macro_decimal_to_f64(&ap_macro.width, variables)
        } else {
            x_end - x_start
        };
        let width = if y_end - y_start == 0. {
            macro_decimal_to_f64(&ap_macro.width, variables)
        } else {
            y_end - y_start
        };

        Self::new(
            Point::new(x_start, y_start)
                + (Point::new(x_end, y_end) - Point::new(x_start, y_start)).div(2.)
                + target,
            height,
            width,
            None,
        )
    }
    pub fn from_macro_center_line(
        ap_macro: &CenterLinePrimitive,
        variables: &[f64],
        target: Point,
    ) -> Self {
        // TODO angle
        Self::new(
            Point::new(
                macro_decimal_to_f64(&ap_macro.center.0, variables),
                macro_decimal_to_f64(&ap_macro.center.1, variables),
            ) + target,
            macro_decimal_to_f64(&ap_macro.dimensions.0, variables),
            macro_decimal_to_f64(&ap_macro.dimensions.1, variables),
            None,
        )
    }

    pub fn into_unit(self, origin: Unit, to: Unit) -> Self {
        Self::new(
            self.position.into_unit(origin, to),
            convert_to_unit(self.width, origin, to),
            convert_to_unit(self.height, origin, to),
            self.hole_diameter.map(|d| convert_to_unit(d, origin, to)),
        )
    }
}

impl DrawableRaw for Rectangle {
    fn canvas_pos(&self) -> Point {
        self.position
            .shift_x(self.position.x() / 2.)
            .shift_y(self.position.y() / 2.)
    }

    fn draw_egui_plot(&self, ui: &mut egui_plot::PlotUi, colour: CanvasColour, selected: bool) {
        let mut points: Vec<[f64; 2]> = self.outline.iter().map(|p| [p.x(), p.y()]).collect();
        points.push(points[0]);

        let poly = Polygon::new(PlotPoints::from(points))
            .fill_color(colour.as_colour32(selected))
            .stroke(Stroke::NONE);
        ui.polygon(poly);
    }

    fn to_paths(&self) -> ClipperPaths {
        ClipperPaths::new(vec![self.outline.clone()])
    }

    fn outline(&self) -> ClipperPath {
        self.outline.clone()
    }
}
