use std::{
    fmt,
    marker::PhantomData,
    ops::{Add, Sub},
};

use clipper2::PointScaler;
use eframe::egui::{Pos2, Vec2};
use egui_plot::PlotPoint;
use gerber_types::Coordinates;

use super::{ClipperPoint, Micro, Unit};

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct Point<P: PointScaler = Micro> {
    x: i64,
    y: i64,
    phantom_data: PhantomData<P>,
}

impl fmt::Debug for Point {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("Point")
            .field("x", &self.x())
            .field("y", &self.y())
            .finish()
    }
}

impl<P: PointScaler> Point<P> {
    pub fn new(x: f64, y: f64) -> Self {
        // Point { x, y }
        Self {
            x: P::scale(x) as i64,
            y: P::scale(y) as i64,
            phantom_data: PhantomData,
        }
    }

    pub fn shift_x(&self, shift: f64) -> Self {
        Self {
            x: self.x + P::scale(shift) as i64,
            y: self.y,
            phantom_data: PhantomData,
        }
    }

    pub fn shift_y(&self, shift: f64) -> Self {
        Self {
            x: self.x,
            y: self.y + P::scale(shift) as i64,
            phantom_data: PhantomData,
        }
    }

    pub fn div(&self, divisor: f64) -> Self {
        Self {
            x: (self.x as f64 / divisor) as i64,
            y: (self.y as f64 / divisor) as i64,
            phantom_data: PhantomData,
        }
    }

    pub fn len(&self) -> f64 {
        P::descale(((self.x.pow(2) + self.y.pow(2)) as f64).sqrt())
    }

    pub fn normalize(&self) -> Self {
        Self::new(self.x() / self.len(), self.y() / self.len())
    }

    pub fn invert_y(&self) -> Self {
        Self {
            x: self.x,
            y: -self.y,
            phantom_data: PhantomData,
        }
    }

    pub fn into_unit(self, origin: Unit, to: Unit) -> Self {
        Self::new(
            convert_to_unit(self.x(), origin, to),
            convert_to_unit(self.y(), origin, to),
        )
    }

    /// Returns the x coordinate of the point.
    pub fn x(&self) -> f64 {
        P::descale(self.x as f64)
    }

    /// Returns the y coordinate of the point.
    pub fn y(&self) -> f64 {
        P::descale(self.y as f64)
    }
}

impl From<Point> for (f64, f64) {
    fn from(value: Point) -> Self {
        (value.x(), value.y())
    }
}

impl From<Point> for [f64; 2] {
    fn from(value: Point) -> Self {
        [value.x(), value.y()]
    }
}

impl From<[f64; 2]> for Point {
    fn from(value: [f64; 2]) -> Self {
        Self::new(value[0], value[1])
    }
}

impl Add for Point {
    type Output = Self;

    fn add(self, other: Self) -> Self {
        Self {
            x: self.x + other.x,
            y: self.y + other.y,
            phantom_data: PhantomData,
        }
    }
}

impl Sub for Point {
    type Output = Self;

    fn sub(self, other: Self) -> Self::Output {
        Self {
            x: self.x - other.x,
            y: self.y - other.y,
            phantom_data: PhantomData,
        }
    }
}

impl From<Point> for Pos2 {
    fn from(value: Point) -> Self {
        Self {
            x: value.x() as f32,
            y: value.y() as f32,
        }
    }
}

impl From<&Point> for Pos2 {
    fn from(value: &Point) -> Self {
        Self {
            x: value.x() as f32,
            y: value.y() as f32,
        }
    }
}

impl From<Pos2> for Point {
    fn from(value: Pos2) -> Self {
        Self::new(value.x.into(), value.y.into())
    }
}

impl From<Point> for Vec2 {
    fn from(value: Point) -> Self {
        Self {
            x: value.x() as f32,
            y: value.y() as f32,
        }
    }
}

impl From<&Point> for ClipperPoint {
    fn from(value: &Point) -> Self {
        Self::from_scaled(value.x, value.y)
    }
}

impl From<&ClipperPoint> for Point {
    fn from(value: &ClipperPoint) -> Self {
        Self::new(value.x(), value.y())
    }
}

impl From<Point> for PlotPoint {
    fn from(value: Point) -> Self {
        Self::new(value.x(), value.y())
    }
}

#[derive(Debug)]
pub enum CoordinateConversionError {
    MissingXValue,
    MissingYValue,
    ParsingError(String),
    FormattingError,
}

impl TryFrom<&Coordinates> for Point {
    type Error = CoordinateConversionError;

    fn try_from(value: &Coordinates) -> Result<Self, Self::Error> {
        let x_coordinate = value
            .x
            .ok_or(CoordinateConversionError::MissingXValue)?
            .gerber(&value.format)
            .map_err(|_| CoordinateConversionError::FormattingError)?
            .parse::<f64>()
            .map_err(|e| CoordinateConversionError::ParsingError(e.to_string()))?;
        let y_coordinate = value
            .y
            .ok_or(CoordinateConversionError::MissingYValue)?
            .gerber(&value.format)
            .map_err(|_| CoordinateConversionError::FormattingError)?
            .parse::<f64>()
            .map_err(|e| CoordinateConversionError::ParsingError(e.to_string()))?;

        Ok(Point::new(
            x_coordinate / 10_f64.powi(value.format.decimal as i32),
            y_coordinate / 10_f64.powi(value.format.decimal as i32),
        ))
    }
}

pub fn convert_to_unit(number: f64, from: Unit, to: Unit) -> f64 {
    match (from, to) {
        (Unit::Millimeters, Unit::Millimeters) | (Unit::Inches, Unit::Inches) => number,
        (Unit::Millimeters, Unit::Inches) => number * 1. / 25.4,
        (Unit::Inches, Unit::Millimeters) => number * 25.4,
    }
}
