use std::collections::HashMap;

use clipper2::{FillRule, PointInPolygonResult};
use itertools::Itertools;

use super::{
    elements::{linepath::LinePath, Element},
    point::Point,
    ClipperBounds, ClipperPaths, ClipperPoint, Unit,
};

#[derive(Debug, Clone, Default)]
pub struct UnitedNets {
    pub conductors: Vec<ConductorNet>,
    pub isolated_apertures: Vec<Element>,
}

impl UnitedNets {
    pub fn into_unit(self, origin: Unit, to: Unit) -> Self {
        Self {
            conductors: self
                .conductors
                .iter()
                .map(|net| net.to_unit(origin, to))
                .collect(),
            isolated_apertures: self
                .isolated_apertures
                .into_iter()
                .map(|el| el.into_unit(origin, to))
                .collect(),
        }
    }

    pub fn bounds(&self) -> ClipperBounds {
        let conductor_net = self
            .conductors
            .iter()
            .map(|net| net.outline.clone())
            .reduce(|mut acc, v| {
                acc.push(v);
                acc
            });
        let isolated_apertures_bound = ClipperPaths::new(
            self.isolated_apertures
                .iter()
                .map(|net| net.outline())
                .collect(),
        )
        .bounds();

        if let Some(net) = conductor_net {
            let net_bounds = net.bounds();

            ClipperBounds {
                min: ClipperPoint::new(
                    net_bounds.min.x().min(isolated_apertures_bound.min.x()),
                    net_bounds.min.y().min(isolated_apertures_bound.min.y()),
                ),
                max: ClipperPoint::new(
                    net_bounds.max.x().max(isolated_apertures_bound.max.x()),
                    net_bounds.max.y().max(isolated_apertures_bound.max.y()),
                ),
            }
        } else {
            isolated_apertures_bound
        }
    }
}

#[derive(Debug, Clone)]
pub struct ConductorNet {
    pub outline: ClipperPaths,
    pub included_points: Vec<Point>,
}

impl ConductorNet {
    pub fn to_unit(&self, origin: Unit, to: Unit) -> Self {
        Self {
            outline: self
                .outline
                .iter()
                .map(|p| {
                    p.iter()
                        .map(|p| (&Point::from(p).into_unit(origin, to)).into())
                        .collect()
                })
                .collect(),
            included_points: self
                .included_points
                .iter()
                .map(|el| el.into_unit(origin, to))
                .collect(),
        }
    }
}

pub fn union_lines(lines: &[LinePath]) -> Vec<ConductorNet> {
    let mut intersection_map = HashMap::new();

    // TODO prevent double checking same combination
    let combinations = lines.iter().enumerate().combinations(2);

    for co in combinations {
        let (index, line) = co[0];
        // print!("{} mit {};", combi[0].0, combi[1].0);
    }

    for (index, line) in lines.iter().enumerate() {
        if !line.is_empty() {
            // create list of intersecting lines
            let mut intersections = Vec::new();
            let (p1, q1) = (line.points[0], line.points[1]);
            // println!("LINE 1 {p1:?}, {q1:?}");
            for (i, l) in lines.iter().enumerate() {
                if !l.is_empty() {
                    // do not check for intersection with itself
                    if index != i {
                        // check for all lines in path
                        // let intersect = l.points.windows(2).any(|w| {
                        //     let (p2, q2) = (w[0], w[1]);
                        //     // println!("LINE 2 {p2:?}, {q2:?}");
                        //     // do_intersect(p1, q1, p2, q2)
                        //     line.outline
                        //         .get(0)
                        //         .unwrap()
                        //         .is_point_inside(ClipperPoint::from(&p2))
                        //         != PointInPolygonResult::IsOutside
                        //         || line
                        //             .outline
                        //             .get(0)
                        //             .unwrap()
                        //             .is_point_inside(ClipperPoint::from(&q2))
                        //             != PointInPolygonResult::IsOutside
                        // });
                        let intersect = line.outline_includes_points(&l.points);
                        // println!("INTERSECTING: {intersect}");
                        if intersect {
                            // let entry = intersection_map.entry(index).or_insert(Vec::new());
                            // entry.push(i);
                            // println!("INTERSECTING {:?} and {:?}", line.points, l.points);
                            intersections.push(i);
                        }
                    }
                }
            }
            //
            intersection_map.insert(index, intersections);
        }
    }

    let mut final_geo = Vec::new();

    // go through all line segments
    for i in 0..intersection_map.len() {
        if let Some(mut intersections) = intersection_map.remove(&i) {
            // get current line path
            let mut geo = lines[i].outline.clone();
            let mut included_points = lines[i].points.clone();

            // union with intersecting lines until done
            while let Some(other) = intersections.pop() {
                // union with intersecting line
                let intersecting_line = &lines[other];
                if let Some(union) = union_function(&geo, &intersecting_line.outline) {
                    geo = union;
                }
                // add points of added line to included points
                for point in &intersecting_line.points {
                    if !included_points.contains(point) {
                        included_points.push(point.to_owned()); // show included points
                    }
                }

                // get intersections of united line and add them if not already in own
                if let Some(new_intersections) = intersection_map.remove(&other) {
                    for o in new_intersections {
                        // add to original line if not self and not already inside
                        if o != i   // ensure to not intersect with itself
                            && !intersections.contains(&o)  // do not add if already in intersection list
                            && intersection_map.contains_key(&o)
                        // check if intersection was already performed
                        {
                            intersections.push(o);
                        }
                    }
                }
            }
            // create conductor net
            final_geo.push(ConductorNet {
                outline: geo,
                included_points,
            })
        }
    }

    final_geo
}

pub fn union_with_apertures(
    apertures: &Vec<Element>,
    conductors: Vec<ConductorNet>,
) -> Option<UnitedNets> {
    let mut isolated_apertures = Vec::new();

    // let mut finalized_paths = Vec::new(); // handle apertures without connection
    let mut current_conductors = conductors;

    // go through all apertures
    for ap in apertures {
        // get outline path
        let geo = ap.to_paths();
        // create empty set of conductors
        let mut new_conductors = Vec::new();
        // create indicator if aperture is isolated
        let mut isolated = true;

        // go through all available conductor nets
        for conductor in current_conductors {
            // check if any point of the conductor net is not outside of the aperture
            if conductor
                .included_points
                .iter()
                .any(|c| ap.outline().is_point_inside(c.into()) != PointInPolygonResult::IsOutside)
            {
                // union aperture with conductor net
                let geo = union_function(&geo, &conductor.outline)?;
                let mut cond = conductor;
                cond.outline = geo;
                isolated = false;
                new_conductors.push(cond);
            } else {
                new_conductors.push(conductor);
            }
        }

        // add aperture to extra container if isolated
        if isolated {
            // finalized_paths.push(geo);
            isolated_apertures.push(ap.clone());
        }

        // update current conductors
        current_conductors = new_conductors;
    }

    Some(UnitedNets {
        conductors: current_conductors,
        isolated_apertures,
    })
}

pub fn union_function(path1: &ClipperPaths, path2: &ClipperPaths) -> Option<ClipperPaths> {
    path1
        .to_clipper_subject()
        .add_clip(path2.clone())
        .union(FillRule::default())
        .ok()
}
