use std::str::Chars;

use gerber_types::{
    ApertureMacro, CenterLinePrimitive, CirclePrimitive, MacroContent, MacroDecimal,
    OutlinePrimitive, PolygonPrimitive, VectorLinePrimitive,
};

use super::doc::GerberDoc;

pub fn start_macro(doc: &mut GerberDoc, macro_lines: &mut Option<Vec<String>>, line: &str) {
    *macro_lines = Some(vec![line.replacen("%AM", "", 1)]);
    if line.ends_with("*%") {
        if let Some(aperture_macro) = parse(macro_lines.as_ref()) {
            doc.aperture_macros
                .insert(aperture_macro.name.to_string(), aperture_macro);
        };
        *macro_lines = None;
    }
}

pub fn continue_macro(doc: &mut GerberDoc, macro_lines: &mut Option<Vec<String>>, line: &str) {
    if let Some(lines) = macro_lines {
        lines.push(line.to_string());
        if line.ends_with("*%") {
            if let Some(aperture_macro) = parse(macro_lines.as_ref()) {
                doc.aperture_macros
                    .insert(aperture_macro.name.to_string(), aperture_macro);
            };
            *macro_lines = None;
        }
    } else {
        start_macro(doc, macro_lines, line);
    }
}

pub fn parse(data: Option<&Vec<String>>) -> Option<ApertureMacro> {
    let mut name = "";
    let mut macro_content = Vec::new();
    let lines = data?;

    // Go through every very part in the macro
    for (index, line) in lines.iter().enumerate() {
        // get name
        if index == 0 {
            name = line.strip_suffix("*")?;
            continue;
        }

        let mut chars = line.chars();
        // remove % char of last line
        if index == lines.len() - 1 {
            chars.next_back();
        }

        match chars.next()? {
            // Comment
            '0' => macro_content.push(MacroContent::Comment(line.to_string())),
            // Circle
            '1' => {
                let args = get_attr_args(chars)?;

                macro_content.push(MacroContent::Circle(CirclePrimitive {
                    exposure: args.first()? == &MacroDecimal::Value(1.0),
                    diameter: args.get(1)?.clone(),
                    center: (args.get(2)?.clone(), args.get(3)?.clone()),
                    angle: match args.get(4) {
                        Some(arg) => Some(arg.clone()),
                        None => Some(MacroDecimal::Value(0.)),
                    },
                }));
            }
            '2' => {
                match chars.next()? {
                    // Vector Line
                    '0' => {
                        let args = get_attr_args(chars)?;

                        macro_content.push(MacroContent::VectorLine(VectorLinePrimitive {
                            exposure: args.first()? == &MacroDecimal::Value(1.0),
                            end: (args.get(4)?.clone(), args.get(5)?.clone()),
                            width: args.get(1)?.clone(),
                            start: (args.get(2)?.clone(), args.get(3)?.clone()),
                            angle: match args.get(6) {
                                Some(arg) => arg.clone(),
                                None => MacroDecimal::Value(0.),
                            },
                        }));
                    }
                    // Center Line
                    '1' => {
                        let args = get_attr_args(chars)?;

                        macro_content.push(MacroContent::CenterLine(CenterLinePrimitive {
                            exposure: args.first()? == &MacroDecimal::Value(1.0),
                            dimensions: (args.get(1)?.clone(), args.get(2)?.clone()),
                            center: (args.get(3)?.clone(), args.get(4)?.clone()),
                            angle: match args.get(6) {
                                Some(arg) => arg.clone(),
                                None => MacroDecimal::Value(0.),
                            },
                        }));
                    }
                    _ => panic!("Unknown command in macro line:\n{} | {}", index, line),
                }
            }
            // Outline
            '4' => {
                let args = get_attr_args(chars)?;

                macro_content.push(MacroContent::Outline(OutlinePrimitive {
                    exposure: args.first()? == &MacroDecimal::Value(1.0),
                    points: (args)
                        .chunks_exact(2)
                        .map(|c| (c[0].clone(), c[1].clone()))
                        .collect(),
                    angle: args.last()?.clone(),
                }));
            }
            // Polygon
            '5' => {
                let args = get_attr_args(chars)?;

                macro_content.push(MacroContent::Polygon(PolygonPrimitive {
                    exposure: args.first()? == &MacroDecimal::Value(1.0),
                    vertices: match args.get(1)? {
                        MacroDecimal::Value(v) => *v as u8,
                        _ => 0,
                    },
                    center: (args.get(2)?.clone(), args.get(3)?.clone()),
                    diameter: args.get(4)?.clone(),
                    angle: match args.get(5) {
                        Some(arg) => arg.clone(),
                        None => MacroDecimal::Value(0.),
                    },
                }));
            }
            '6' => todo!(), // Moiré
            '7' => todo!(), // Thermal
            '$' => todo!(), // Define variable
            _ => panic!("Unknown command in macro line:\n{} | {}", index, line),
        }
    }

    Some(ApertureMacro {
        name: name.to_string(),
        content: macro_content,
    })
}

fn get_attr_args(mut attribute_chars: Chars) -> Option<Vec<MacroDecimal>> {
    attribute_chars.next_back()?;
    attribute_chars.next()?;

    Some(
        attribute_chars
            .as_str()
            .split(",")
            .map(|el| el.trim())
            .map(|el| {
                if el.starts_with("$") {
                    MacroDecimal::Expression(el.to_string())
                } else {
                    MacroDecimal::Value(el.parse::<f64>().unwrap())
                }
            })
            .collect(),
    )
}

// def parse_content(self):
//         """
//         Creates numerical lists for all primitives in the aperture
//         macro (in ``self.raw``) by replacing all variables by their
//         values iteratively and evaluating expressions. Results
//         are stored in ``self.primitives``.

//         :return: None
//         """
//         # Cleanup
//         self.raw = self.raw.replace('\n', '').replace('\r', '').strip(" *")
//         self.primitives = []

//         # Separate parts
//         parts = self.raw.split('*')

//         #### Every part in the macro ####
//         for part in parts:
//             ### Comments. Ignored.
//             match = ApertureMacro.amcomm_re.search(part)
//             if match:
//                 continue

//             ### Variables
//             # These are variables defined locally inside the macro. They can be
//             # numerical constant or defind in terms of previously define
//             # variables, which can be defined locally or in an aperture
//             # definition. All replacements ocurr here.
//             match = ApertureMacro.amvar_re.search(part)
//             if match:
//                 var = match.group(1)
//                 val = match.group(2)

//                 # Replace variables in value
//                 for v in self.locvars:
//                     val = re.sub(r'\$'+str(v)+r'(?![0-9a-zA-Z])', str(self.locvars[v]), val)

//                 # Make all others 0
//                 val = re.sub(r'\$[0-9a-zA-Z](?![0-9a-zA-Z])', "0", val)

//                 # Change x with *
//                 val = re.sub(r'[xX]', "*", val)

//                 # Eval() and store.
//                 self.locvars[var] = eval(val)
//                 continue

//             ### Primitives
//             # Each is an array. The first identifies the primitive, while the
//             # rest depend on the primitive. All are strings representing a
//             # number and may contain variable definition. The values of these
//             # variables are defined in an aperture definition.
//             match = ApertureMacro.amprim_re.search(part)
//             if match:
//                 ## Replace all variables
//                 for v in self.locvars:
//                     part = re.sub(r'\$' + str(v) + r'(?![0-9a-zA-Z])', str(self.locvars[v]), part)

//                 # Make all others 0
//                 part = re.sub(r'\$[0-9a-zA-Z](?![0-9a-zA-Z])', "0", part)

//                 # Change x with *
//                 part = re.sub(r'[xX]', "*", part)

//                 ## Store
//                 elements = part.split(",")
//                 self.primitives.append([eval(x) for x in elements])
//                 continue

//             log.warning("Unknown syntax of aperture macro part: %s" % str(part))
