use clipper2::{EndType, JoinType};

use crate::{
    excellon::drills::Drills,
    geometry::{
        point::Point, union::UnitedNets, ClipperBounds, ClipperPath, ClipperPaths, DrawableRaw,
        Unit,
    },
};

#[derive(Debug, Clone)]
pub enum GeometryType {
    Paths(ClipperPaths),
    Points(Vec<Point>),
}

#[derive(Debug, Clone)]
pub struct OutlineGeometry {
    items: GeometryType,
    pub stroke: f32,
    pub unit: Unit,
    pub bounds_from: String,
    pub bounding_box: ClipperBounds,
}

impl OutlineGeometry {
    pub fn new(united_nets: &UnitedNets, stroke: f32, unit: Unit, bounds_from: &str) -> Self {
        let mut outline_paths = ClipperPaths::new(vec![]);
        // inflate conductor net paths
        for net in &united_nets.conductors {
            outline_paths.push(
                net.outline
                    .inflate((stroke / 2.).into(), JoinType::Round, EndType::Polygon, 2.)
                    .simplify(0.001, false),
            )
        }
        // inflate isolated apertures
        for ap in &united_nets.isolated_apertures {
            outline_paths.push(
                ap.to_paths()
                    .inflate((stroke / 2.).into(), JoinType::Round, EndType::Polygon, 2.)
                    .simplify(0.001, false),
            );
        }

        Self {
            items: GeometryType::Paths(outline_paths),
            stroke,
            unit,
            bounds_from: bounds_from.into(),
            bounding_box: united_nets.bounds(),
        }
    }

    pub fn new_no_inflate(
        outline: &ClipperPaths,
        stroke: f32,
        unit: Unit,
        bounds_from: &str,
    ) -> Self {
        Self {
            items: GeometryType::Paths(outline.clone()),
            stroke,
            unit,
            bounds_from: bounds_from.into(),
            bounding_box: outline.bounds(),
        }
    }

    pub fn drill_marker(drills: &Drills, stroke: f32, unit: Unit, bounds_from: &str) -> Self {
        Self {
            items: GeometryType::Points(drills.holes.iter().map(|c| c.canvas_pos()).collect()),
            stroke,
            unit,
            bounds_from: bounds_from.into(),
            bounding_box: ClipperPaths::from(
                drills
                    .holes
                    .iter()
                    .map(|hole| hole.outline.clone())
                    .collect::<Vec<ClipperPath>>(),
            )
            .bounds(),
        }
    }

    pub fn paths(&self) -> ClipperPaths {
        match &self.items {
            GeometryType::Paths(paths) => paths.clone(),
            GeometryType::Points(_) => ClipperPaths::new(vec![]),
        }
    }

    pub fn points(&self) -> Vec<Point> {
        match &self.items {
            GeometryType::Paths(_) => vec![],
            GeometryType::Points(p) => p.clone(),
        }
    }
}
